import numpy as np
import teneto
import collections
import scipy.spatial.distance as distance
from nilearn.input_data import NiftiSpheresMasker, NiftiLabelsMasker
from nilearn.datasets import fetch_atlas_harvard_oxford
import json

"""

Couple of utiltity functions for teneto for converting between graphlet and contact sequence representations

"""


def graphlet2contact(G, params=None):
    """

    Converts graphlet (snapshot) representation of temporal network and converts it to contact representation representation of network. Contact representation are more efficient for memory storing. Also includes metadata which can made it easier for plotting. A contact representation contains all non-zero edges.

    Parameters
    ----------
    G : array_like
        Temporal network.
    params : dict, optional
        Dictionary of parameters for contact representation.

        *Fs* : int, default=1
            sampling rate.

        *timeunit* : str, default=''
            Sampling rate in for units (e.g. seconds, minutes, years).

        *nettype* : str, default='auto'
            Define what type of network. Can be:
            'auto': detects automatically;
            'wd': weighted, directed;
            'bd': binary, directed;
            'wu': weighted, undirected;
            'bu': binary, undirected.

        *diagonal* : int, default = 0.
            What should the diagonal be. (note: does could be expanded to take vector of unique diagonal values in the future, but not implemented now)

        *timetype* : str, default='discrete'
            Time units can be The params file becomes the foundation of 'C'. Any other information in params, will added to C.

        *nLabs* : list
            Set nod labels.

        *t0*: int
            Time label at first index.


    Returns
    -------

    C : dict

        Contact representation of temporal network.
        Includes 'contacts', 'values' (if nettype[0]='w'),'nettype','netshape', 'Fs', 'dimord' and 'timeunit', 'timetype'.

    """

    # Create config dictionary if missing
    if params == None:
        params = {}
    # Check that temporal network is vald input.
    if G.shape[0] != G.shape[1]:
        raise ValueError(
            'Input G (node x node x time), requires Rows and Columns to be the same size.')
    if len(G.shape) == 2:
        G = np.atleast_3d(G)
    if len(G.shape) != 3:
        raise ValueError(
            'Input G must be three dimensions (node x node x time)')
    # Check number of nodes is correct, if specfied
    if 'nLabs' in params.keys():
        if params['nLabs']:
            if len(params['nLabs']) != G.shape[0]:
                raise ValueError(
                    'Specified list of node names has to be equal in length to number of nodes')
    if 't0' in params.keys():
        params['t0'] = np.atleast_1d(np.array(params['t0']))
        if len(params['t0']) != 1:
            raise ValueError(
                't0 must be sigular be either integer representing time at first temporal index)')
        params['t0'] = np.squeeze(params['t0'])
    # Check that all inputs in params are correct.

    if 'nettype' not in params.keys() or params['nettype'] == 'auto':
        params['nettype'] = gen_nettype(G, 1)
    if params['nettype'] not in {'bd', 'bu', 'wd', 'wu', 'auto'}:
        raise ValueError(
            '\'nettype\' (in params) must be a string \'wd\',\'bd\',\'wu\',\'bu\'). w: weighted network. b: binary network. u: undirected network. d: directed network')
    if 'Fs' not in params.keys():
        params['Fs'] = 1
        print('Warning, no sampling rate set. Assuming 1.')
    if 'timeunit' not in params.keys():
        params['timeunit'] = ''
    if 'diagonal' not in params.keys():
        params['diagonal'] = 0
    if 'nLabs' not in params.keys():
        params['nLabs'] = ''
    else:
        params['nLabs'] = list(params['nLabs'])

    if 't0' not in params.keys():
        params['t0'] = 1
    nt = params['nettype']

    # Set diagonal to 0 to make contacts 0.
    G = set_diagonal(G, 0)

    # Very convoluted way to get all the indexes into a tuple, ordered by time
    if nt[1] == 'u':
        G = [np.triu(G[:, :, t], k=1) for t in range(0, G.shape[2])]
        G = np.transpose(G, [1, 2, 0])
    edg = np.where(np.abs(G) > 0)
    sortTime = np.argsort(edg[2])
    contacts = np.array([tuple([edg[0][i], edg[1][i], edg[2][i]])
                         for i in sortTime])
    # Get each of the values if weighted matrix
    if nt[0] == 'w':
        values = list(G[edg[0][sortTime], edg[1][sortTime], edg[2][sortTime]])

    # build output dictionary
    C = params
    C['contacts'] = contacts
    C['netshape'] = G.shape
    C['dimord'] = 'node,node,time'
    # Obviously this needs to change
    C['timetype'] = 'discrete'
    if nt[0] == 'w':
        C['values'] = values

    return C


def contact2graphlet(C):
    """

    Converts contact representation to graphlet (snaptshot) representation. Graphlet representation discards all meta information in the contact representation.

    Parameters
    ----------

    C : dict
        A contact representation. Must include keys: 'dimord', 'netshape', 'nettype', 'contacts' and, if weighted, 'values'.

    Returns
    -------

    G: array
        Graphlet representation of temporal network.

    Note
    ----

    Returning elements of G will be float, even if binary graph.

    """

    # Check that contact sequence is vald input.
    if 'dimord' not in C.keys():
        raise ValueError('\'dimord\' must be present in C.')
    if C['dimord'] != 'node,node,time':
        raise ValueError('\'dimord\' must be string \'node,node,time\'.')
    if 'dimord' not in C.keys():
        raise ValueError('\'dimord\' must be present in C.')
    if C['dimord'] != 'node,node,time':
        raise ValueError('\'dimord\' must be string \'node,node,time\'.')
    if 'nettype' not in C.keys():
        raise ValueError(
            'C must include parameter \'nettype\' (wd,bd,wu,bu). w: weighted network. b: binary network. u: undirected network. d: directed network')
    if C['nettype'] not in {'bd', 'bu', 'wd', 'wu'}:
        raise ValueError(
            '\'nettype\' in (C) must be a string \'wd\',\'bd\',\'wu\',\'bu\'). w: weighted network. b: binary network. u: undirected network. d: directed network')
    if 'netshape' not in C.keys():
        raise ValueError(
            'C must include netshape expressing size of target network (tuple)')
    if not isinstance(C['netshape'], tuple):
        raise ValueError('\'netshape\' (in C) should be a tuple')
    if len(C['netshape']) != 3:
        raise ValueError('\'netshape\' tuple should be of 3 dimensions')
    if C['nettype'][0] == 'w' and 'values' not in C.keys():
        raise ValueError('values not in C and asked for weighted network')
    if 'contacts' not in C.keys():
        raise ValueError('contacts must be expressed (list of tuples)')
    if C['timetype'] != 'discrete':
        print('Warning: timetype is not discrete. In future updates timetype in dictionary should be \'discrete\' to be converted to grpahlets')

    nt = C['nettype']

    # Preallocate
    G = np.zeros(C['netshape'])

    # Convert indexes of C to numpy friend idx list
    idx = np.array(list(map(list, C['contacts'])))
    if nt[0] == 'b':
        G[idx[:, 0], idx[:, 1], idx[:, 2]] = 1
        if nt[1] == 'u':
            G[idx[:, 1], idx[:, 0], idx[:, 2]] = 1
    elif nt[0] == 'w':
        G[idx[:, 0], idx[:, 1], idx[:, 2]] = C['values']
        if nt[1] == 'u':
            G[idx[:, 1], idx[:, 0], idx[:, 2]] = C['values']
    # If diagonal is not 0, fill it to whatever it is set to
    if C['diagonal'] != 0:
        G = set_diagonal(G, C['diagonal'])

    return G


def binarize_percent(netin, level, sign='pos', axis='time'):
    """
    Binarizes a network proprtionally. When axis='time' (only one available at the moment) then the top values for each edge time series are considered.

    Parameters
    ----------

    netin : array or dict
        network (graphlet or contact representation),
    level : float
        Percent to keep (expressed as decimal, e.g. 0.1 = top 10%)
    sign : str, default='pos'
        States the sign of the thresholding. Can be 'pos', 'neg' or 'both'. If "neg", only negative values are thresholded and vice versa.
    axis : str, default='time'
        Specify which dimension thresholding is applied against. Only 'time' option exists at present.

    Returns
    -------

    netout : array or dict (depending on input)
        Binarized network

    """
    netin, netinfo = teneto.utils.process_input(netin, ['C', 'G', 'TO'])
    if sign == 'both':
        net_sorted = np.argsort(np.abs(netin),axis=-1)
    elif sign == 'pos':
        net_sorted = np.argsort(netin,axis=-1)
    elif sign == 'neg':
        net_sorted = np.argsort(-1*netin,axis=-1)
    else:
        raise ValueError('Unknown value for parameter: sign')
    # Predefine
    netout = np.zeros(netinfo['netshape'])
    # These for loops can probabaly be removed for speed
    for i in range(netinfo['netshape'][0]):
        for j in range(netinfo['netshape'][1]):
            netout[i,j,net_sorted[i,j,-int(round(net_sorted.shape[-1])*level):]] = 1
    # Set diagonal to 0
    netout = teneto.utils.set_diagonal(netout,0)

    # If input is contact, output contact
    if netinfo['inputtype'] == 'C':
        netinfo['nettype'] = 'b' + netinfo['nettype'][1]
        netout = teneto.utils.graphlet2contact(netout,netinfo)
        netout.pop('inputtype')
        netout.pop('values')
        netout['diagonal'] = 0

    return netout


# To do: set diagonal to 0.
def binarize_rdp(netin, level, sign='pos', axis='time'):
    """
    Binarizes a network based on RDP compression.

    Parameters
    ----------

    netin : array or dict
        Network (graphlet or contact representation),
    level : float
        Delta parameter which is the tolorated error in RDP compression.
    sign : str, default='pos'
        States the sign of the thresholding. Can be 'pos', 'neg' or 'both'. If "neg", only negative values are thresholded and vice versa.

    Returns
    -------

    netout : array or dict (dependning on input)
        Binarized network
    """
    netin, netinfo = teneto.utils.process_input(netin, ['C', 'G', 'TO'])
    trajectory = teneto.trajectory.rdp(netin,level)

    contacts = []
    # Use the trajectory points as threshold
    for n in range(trajectory['index'].shape[0]):
        if sign == 'pos':
            sel = trajectory['trajectory_points'][n][trajectory['trajectory'][n][trajectory['trajectory_points'][n]]>0]
        elif sign == 'neg':
            sel = trajectory['trajectory_points'][n][trajectory['trajectory'][n][trajectory['trajectory_points'][n]]<0]
        else:
            sel = trajectory['trajectory_points']
        i_ind = np.repeat(trajectory['index'][n,0],len(sel))
        j_ind = np.repeat(trajectory['index'][n,1],len(sel))
        contacts.append(np.array([i_ind,j_ind,sel]).transpose())
    contacts = np.concatenate(contacts)

    # Create output dictionary
    netout = dict(netinfo)
    netout['contacts'] = contacts
    netout['nettype'] = 'b' + netout['nettype'][1]
    netout['dimord'] = 'node,node,time'
    netout['timetype'] = 'discrete'
    netout['diagonal'] = 0
    # If input is graphlet, output graphlet
    if netinfo['inputtype'] == 'G':
        netout = teneto.utils.contact2graphlet(netout)
    else:
        netout.pop('inputtype')

    return netout

def binarize_magnitude(netin, level, sign='pos'):
    """

    Parameters
    ----------

    netin : array or dict
        Network (graphlet or contact representation),
    level : float
        Magnitude level threshold at.
    sign : str, default='pos'
        States the sign of the thresholding. Can be 'pos', 'neg' or 'both'. If "neg", only negative values are thresholded and vice versa.
    axis : str, default='time'
        Specify which dimension thresholding is applied against. Only 'time' option exists at present.

    Returns
    -------

    netout : array or dict (depending on input)
        Binarized network
    """
    netin, netinfo = teneto.utils.process_input(netin, ['C', 'G', 'TO'])
    # Predefine
    netout = np.zeros(netinfo['netshape'])

    if sign == 'pos' or sign == 'both':
        netout[netin>level] = 1
    if sign == 'neg' or sign == 'both':
        netout[netin<level] = 1

    # Set diagonal to 0
    netout = teneto.utils.set_diagonal(netout,0)

    # If input is contact, output contact
    if netinfo['inputtype'] == 'C':
        netinfo['nettype'] = 'b' + netinfo['nettype'][1]
        netout = teneto.utils.graphlet2contact(netout,netinfo)
        netout.pop('inputtype')
        netout.pop('values')
        netout['diagonal'] = 0

    return netout

def binarize(netin, threshold_type, threshold_level, sign='pos'):
    """
    Binarizes a network, returning the network. General wrapper function for different binarization functions.

    Parameters
    ----------

    netin : array or dict
       Network (graphlet or contact representation),

    threshold_type : str
        What type of thresholds to make binarization. Options: 'rdp', 'percent', 'magnitude'.

    threshold_level : str
        Paramter dependent on threshold type.
        If 'rdp', it is the delta (i.e. error allowed in compression).
        If 'percent', it is the percentage to keep (e.g. 0.1, means keep 10% of signal).
        If 'magnitude', it is the amplitude of signal to keep.

    sign : str, default='pos'
        States the sign of the thresholding. Can be 'pos', 'neg' or 'both'. If "neg", only negative values are thresholded and vice versa.

    Returns
    -------

    netout : array or dict (depending on input)
        Binarized network

    """
    if threshold_type == 'percent':
        netout = teneto.utils.binarize_percent(netin,threshold_level,sign)
    elif threshold_type == 'magnitude':
        netout = teneto.utils.binarize_magnitude(netin,threshold_level,sign)
    elif threshold_type == 'rdp':
        netout = teneto.utils.binarize_rdp(netin,threshold_level,sign)
    else:
        raise ValueError('Unknown value to parameter: threshold_type.')
    return netout

def set_diagonal(G, val=0):
    """

    Generally diagonal is set to 0. This function helps set the diagonal across time.

    Parameters
    ----------

    G : array
        temporal network (graphlet)
    val : value to set diagonal to (default 0).

    Returns
    -------

    G : array
        Graphlet representation with new diagonal

    """

    for t in range(0, G.shape[2]):
        np.fill_diagonal(G[:, :, t], val)
    return G


def gen_nettype(G, printWarning=0):
    """

    Attempts to identify what nettype input graphlet G is. Diagonal is ignored.

    Paramters
    ---------

    G : array
        temporal network (graphlet)

    printWarning : int, default=0
        Options: 0 (default) or 1. Prints warning in console so user knows assumptions made about inputted data.

    """

    if set(np.unique(G)) == set([0, 1]):
        weights = 'b'
    else:
        weights = 'w'

    if np.allclose(G.transpose(1, 0, 2), G):
        direction = 'u'
    else:
        direction = 'd'

    nettype = weights + direction

    if printWarning == 1:
        netNames = {'w': 'weighted', 'b': 'binary',
                    'u': 'undirected', 'd': 'directed'}
        print('Assuming network is ' +
              netNames[nettype[0]] + ' and ' + netNames[nettype[1]] + '.')

    return nettype


def checkInput(netIn, raiseIfU=1, conMat=0):
    """

    This function checks that netIn input is either graphlet (G) or contact (C).

    Parameters
    ----------

    netIn : array or dict
        temporal network, (graphlet or contact).
    raiseIfU : int, default=1.
        Options 1 or 0. Error is raised if not found to be G or C
    conMat : int, default=0.
        Options 1 or 0. If 1, input is allowed to be a 2 dimensional connectivity matrix. Allows output to be 'M'

    Returns
    -------

    inputtype : str
        String indicating input type. 'G','C', 'M' or 'U' (unknown). M is special case only allowed when conMat=1 for a 2D connectivity matrix.

    """

    inputIs = 'U'
    if isinstance(netIn, np.ndarray):
        netShape = netIn.shape
        if len(netShape) == 3 and netShape[0] == netShape[1]:
            inputIs = 'G'
        if netShape[0] == netShape[1] and conMat == 1:
            inputIs = 'M'

    elif isinstance(netIn, dict):
        if 'nettype' in netIn and 'contacts' in netIn and 'dimord' in netIn and 'timetype' in netIn:
            if netIn['nettype'] in {'bd', 'bu', 'wd', 'wu'} and netIn['timetype'] == 'discrete' and netIn['dimord'] == 'node,node,time':
                inputIs = 'C'

    elif isinstance(netIn, object):
        if isinstance(netIn.contact, dict):
            if 'nettype' in netIn.contact and 'contacts' in netIn.contact and 'dimord' in netIn.contact and 'timetype' in netIn.contact:
                if netIn.contact['nettype'] in {'bd', 'bu', 'wd', 'wu'} and netIn.contact['timetype'] == 'discrete' and netIn.contact['dimord'] == 'node,node,time':
                    inputIs = 'TO'

    if raiseIfU == 1 and inputIs == 'U':
        raise ValueError(
            'Input cannot be identified as graphlet or contact representation')

    return inputIs


def getDistanceFunction(requested_metric):
    """

    This function returns a specified distance function.


    Paramters
    ---------

    requested_metric: str
        Distance function. Can be any function in: https://docs.scipy.org/doc/scipy/reference/spatial.distance.html.

    Returns
    -------

    requested_metric : distance function

    """

    distance_options = {
        'braycurtis': distance.braycurtis,
        'canberra': distance.canberra,
        'chebyshev': distance.chebyshev,
        'cityblock': distance.cityblock,
        'correlation': distance.correlation,
        'cosine': distance.cosine,
        'euclidean': distance.euclidean,
        'sqeuclidean': distance.sqeuclidean,
        'dice': distance.dice,
        'hamming': distance.hamming,
        'jaccard': distance.jaccard,
        'kulsinski': distance.kulsinski,
        'matching': distance.matching,
        'rogerstanimoto': distance.rogerstanimoto,
        'russellrao': distance.russellrao,
        'sokalmichener': distance.sokalmichener,
        'sokalsneath': distance.sokalsneath,
        'yule': distance.yule,
    }

    if requested_metric in distance_options:
        return distance_options[requested_metric]
    else:
        raise ValueError('Distance function cannot be found.')


def process_input(netIn, allowedformats, outputformat='G'):
    """
    Takes input network and checks what the input is.

    Parameters
    ----------

    netIn : array, dict, or class
        Network (graphlet, contact or object)
    allowedformats : str
        Which format of network objects that are allowed. Options: 'C', 'TO', 'G'.
    outputformat: str, default=G
        Target output format. Options: 'C' or 'G'.

    Returns
    -------

    C : dict

    OR

    G : array
        Graphlet representation.
    netInfo : dict
        Metainformation about network.

    """
    inputtype = teneto.utils.checkInput(netIn)
    # Convert TO to C representation
    if inputtype == 'TO' and 'TO' in allowedformats:
        G = netIn.get_graphlet_representation()
        netInfo = dict(netIn.contact)
        netInfo.pop('contacts')
    # Convert C representation to G
    elif inputtype == 'C' and 'C' in allowedformats and outputformat != 'C':
        G = teneto.utils.contact2graphlet(netIn)
        netInfo = dict(netIn)
        netInfo.pop('contacts')
        nettype = netIn['nettype']
    # Get network type if not set yet
    elif inputtype == 'G' and 'G' in allowedformats:
        netInfo = {}
        netInfo['netshape'] = netIn.shape
        netInfo['nettype'] = teneto.utils.gen_nettype(netIn)
        G = netIn
    elif inputtype == 'C' and outputformat == 'C':
        pass
    else:
        raise ValueError('Input invalid.')
    netInfo['inputtype'] = inputtype
    if inputtype != 'C' and outputformat == 'C':
        C = teneto.utils.graphlet2contact(netIn, netInfo)
    if outputformat == 'G':
        return G, netInfo
    elif outputformat == 'C':
        return C


def clean_community_indexes(communityID):
    """
    Takes input of community assignments. Returns reindexed community assignment by using smallest numbers possible.

    Parameters
    ----------

    communityID : array-like
        list or array of integers. Output from community detection algorithems.

    Returns
    -------

    new_communityID : array
        cleaned list going from 0 to len(np.unique(communityID))-1

    Note
    -----

    Behaviour of funciton entails that the lowest community integer in communityID will recieve the lowest integer in new_communityID.

    """
    communityID = np.array(communityID)
    cid_shape = communityID.shape
    if len(cid_shape) > 1: 
        communityID = communityID.flatten()  
    new_communityID = np.zeros(len(communityID))
    for i, n in enumerate(np.unique(communityID)):
        new_communityID[communityID == n] = i
    if len(cid_shape) > 1: 
        new_communityID = new_communityID.reshape(cid_shape)  
    return new_communityID


def multiple_contacts_get_values(C):
    """
    Given an contact representation with repeated contacts, this function removes duplicates and creates a value

    Parameters
    ----------

    C : dict

        contact representation with multiple repeated contacts.

    Returns
    -------

    :C_out: dict

        Contact representation with duplicate contacts removed and the number of duplicates is now in the 'values' field.

    """
    d = collections.OrderedDict()
    for c in C['contacts']:
        ct = tuple(c)
        if ct in d:
            d[ct] += 1
        else:
            d[ct] = 1

    new_contacts = []
    new_values = []
    for (key, value) in d.items():
        new_values.append(value)
        new_contacts.append(key)
    C_out = C
    C_out['contacts'] = new_contacts
    C_out['values'] = new_values
    return C_out



def check_distance_funciton_input(distance_func_name,netinfo):
    """
    Funciton checks distance_func_name, if it is specified as 'default'. Then given the type of the network selects a default distance function.

    Parameters
    ----------

    distance_func_name : str
        distance function name.

    netinfo : dict
        the output of utils.process_input

    Returns
    -------

    distance_func_name : str
        distance function name.
    """

    if distance_func_name == 'default' and netinfo['nettype'][0] == 'b':
        print('Default distance funciton specified. As network is binary, using Hamming')
        distance_func_name = 'hamming'
    elif distance_func_name == 'default' and netinfo['nettype'][0] == 'w':
        distance_func_name = 'euclidean'
        print(
            'Default distance funciton specified. '
            'As network is weighted, using Euclidean')

    return distance_func_name




def load_parcellation_coords(parcellation_name):
    """
    Loads coordinates of included parcellations.

    Parameters
    ----------

    parcellation_name : str
        options: 'gordon2014_333', 'power2012_264', 'shen2013_278'.

    Returns
    -------
    parc : array
        parcellation cordinates

    """

    path = teneto.__path__[0] + '/data/parcellation/' + parcellation_name + '.csv'
    parc = np.loadtxt(path,skiprows=1,delimiter=',',usecols=[1,2,3])

    return parc


def make_parcellation(data_path, parcellation, parc_type=None, parc_params=None):

    """
    Performs a parcellation which reduces voxel space to regions of interest (brain data).

    Parameters
    ----------

    data_path : str
        Path to .nii image.
    parcellation : str
        Specify which parcellation that you would like to use. For MNI: 'gordon2014_333', 'power2012_264', For TAL: 'shen2013_278'.
        It is possible to add the OH subcotical atlas on top of a cortical atlas (e.g. gordon) by adding:
            '+sub-maxprob-thr0-1mm', '+sub-maxprob-thr0-2mm', 'sub-maxprob-thr25-1mm', 'sub-maxprob-thr25-2mm',
            '+sub-maxprob-thr50-1mm', '+sub-maxprob-thr50-2mm'.
            e.g.: gordon2014_333+submaxprob-thr0-2mm'
    parc_type : str
        Can be 'sphere' or 'region'. If nothing is specified, the default for that parcellation will be used.
    parc_params : dict
        **kwargs for nilearn functions

    Returns
    -------

    data : array
        Data after the parcellation.

    NOTE
    ----
    These functions make use of nilearn. Please cite nilearn if used in a publicaiton.
    """

    if isinstance(parcellation,str):

        if '+' in parcellation:
            parcin = parcellation.split('+')
            parcellation = parcin[0]
            subcortical = parcin[1]
        else:
            subcortical = None

        if not parc_type or not parc_params:
            path = teneto.__path__[0] + '/data/parcellation_defaults/defaults.json'
            with open(path) as data_file:
                defaults = json.load(data_file)
        if not parc_type:
            parc_type = defaults[parcellation]['type']
            print('Using default parcellation type')
        if not parc_params:
            parc_params = defaults[parcellation]['params']
            print('Using default parameters')

    if parc_type == 'sphere':
        parcellation = teneto.utils.load_parcellation_coords(parcellation)
        seed = NiftiSpheresMasker(np.array(parcellation),**parc_params)
        data = seed.fit_transform(data_path)
    elif parc_type == 'region':
        path = teneto.__path__[0] + '/data/parcellation/' + parcellation + '.nii'
        region = NiftiLabelsMasker(path,**parc_params)
        data = region.fit_transform(data_path)
    else:
        raise ValueError('Unknown parc_type specified')

    if subcortical:
        subatlas = fetch_atlas_harvard_oxford('sub-maxprob-thr0-2mm')['maps']
        region = NiftiLabelsMasker(subatlas,**parc_params)
        data_sub = region.fit_transform(data_path)
        data = np.hstack([data, data_sub])

    return data



def create_traj_ranges(start, stop, N):

    """
    Fills in the trajectory range.

    # Adapted from https://stackoverflow.com/a/40624614
    """
    steps = (1.0/(N-1)) * (stop - start)
    if np.isscalar(steps):
        return steps*np.arange(N) + start
    else:
        return steps[:,None]*np.arange(N) + start[:,None]



def get_dimord(measure,calc=None,subnet=None):

    """
    Get the dimension order of a network measure.

    Parameters
    ----------

    measure : str
        Name of funciton in teneto.networkmeasures.
    calc : str, default=None
        Calc parameter for the function
    subnet : bool, default=None
        If not null, then subnet property is assumed to be believed.

    Returns
    -------

    dimord : str
        Dimension order. So "node,node,time" would define the dimensions of the network measure.

    """

    if not calc:
        calc = ''
    else:
        calc = '_' + calc
    if not subnet:
        subnet = ''
    else:
        subnet = '_subnet'
    if '_subnet' in calc and '_subnet' in subnet:
        subnet = ''
    if calc == 'subnet_avg' or calc == 'subnet_pairs':
        subnet = ''

    dimord_dict = {
        'temporal_closeness_centrality': 'node',
        'temporal_degree_centrality': 'node',
        'temporal_degree_centralit_avg': 'node',
        'temporal_degree_centrality_time': 'node,time',
        'temporal_degree_centrality_time_subnet': 'subnet,subnet,time',
        'temporal_degree_centrality_subnet': 'subnet,subnet',
        'temporal_degree_centrality_avg_subnet': 'subnet,subnet',
        'temporal_efficiency': 'global',
        'temporal_efficiency_global': 'global',
        'temporal_efficiency_node': 'node',
        'temporal_efficiency_to': 'node',
        'sid_global': 'global,time',
        'sid_global_subnet': 'global,time',
        'sid_subnet_pairs': 'subnet,subnet,time',
        'sid_subnet_avg': 'subnet,time',
        'sid': 'subnet,subnet,time',
        'reachability_latency_global': 'global',
        'reachability_latency': 'global',
        'reachability_latency_node': 'node',
        'fluctuability': 'node',
        'fluctuability_global': 'global',
        'bursty_coeff': 'edge,edge',
        'bursty_coeff_edge': 'edge,edge',
        'bursty_coeff_node': 'node',
        'bursty_coeff_subnet': 'subnet,subnet',
        'bursty_coeff_meanEdgePerNode': 'node',
        'volatility_global': 'time',
    }
    if measure + calc + subnet in dimord_dict:
        return dimord_dict[measure + calc + subnet]
    else:
        print('WARNINGL: get_dimord() returned unknown dimension labels')
        return 'unknown'





import teneto.utils as utils
import numpy as np
from teneto.networkmeasures.temporal_degree_centrality import temporal_degree_centrality


def sid(net, subnet, axis=0, calc='global', decay=None):
    """

    Segregation integration difference (SID). An estimation of each subnetwork (or global) difference of within versus between subnetwork strength.

     **PARAMETERS**

     :net: temporal network input (graphlet or contact).

         :nettype: 'bu', 'bd', 'wu', 'wd'

    :axis: Dimension that is returned 0 or 1 (default 0).
        Note, only relevant for directed networks.
        i.e. if 0, node i has Aijt summed over j and t.
        and if 1, node j has Aijt summed over i and t.

    :calc: 'global' (returns temporal degree cent rality (a 1xnode vector))
      or 'subnet_pairs' (returns a subnetwork x subnetwork x time matrix). Which is the SID for each subnet pairing
      of 'subnet_avg' (returns a subnetwork x time matrix). Which is the normalized average of each subnetwork to all other networks.


    :subnet: a Nx1 vector of subnetwork assignment.

    :decay: if calc = 'time', then decay is possible where the centrality of
    the previous time point is carried over to the next time point but decays
    at a value of $e^decay$ such that the temporal centrality measure becomes: $D_d(t+1) = e^{-decay}D_d(t) + D(t+1)$.

    **OUTPUT**

    :sid: segregation-integration difference

        :format: 2d or 3d numpy array (depending on calc) representing (subnet,subnet,time) or (subnet,time)

    **SEE ALSO**

    - *temporal_degree_centrality*

    **Source**

    Fransson et al (2018) Brain network segregation and integration during an epoch-related working memory fMRI experiment.
    https://www.biorxiv.org/content/early/2018/01/23/252338

    """

    net, netinfo = utils.process_input(net, ['C', 'G', 'TO'])
    D = temporal_degree_centrality(net, calc='time', subnet=subnet, decay=decay)
    # Check network output (order of subnetworks)
    network_ids = np.unique(subnet)
    subnet_size = np.array([sum(subnet==n) for n in network_ids])

    sid = np.zeros([network_ids.max()+1,network_ids.max()+1,net.shape[-1]])
    for n in network_ids:
        for m in network_ids:
            betweenmodulescaling = 1/(subnet_size[n]*subnet_size[m])
            if netinfo['nettype'][1] == 'd':
                withinmodulescaling = 1/(subnet_size[n]*subnet_size[n])
            elif netinfo['nettype'][1] == 'u':
                withinmodulescaling = 2/(subnet_size[n]*(subnet_size[n]-1))
                if n == m:
                    betweenmodulescaling = withinmodulescaling
            sid[n,m,:] = withinmodulescaling * D[n,n,:] - betweenmodulescaling * D[n,m,:]

    if calc == 'global':
        return np.sum(np.sum(sid,axis=1),axis=0)
    elif calc == 'subnet_avg':
        return np.sum(sid,axis=axis)
    else:
        return sid



"""
derive.report
"""

import os
import matplotlib.pyplot as plt
import numpy as np


def gen_report(report, sdir='./', report_name = 'report.html'):

    """
    Generates report of derivation and postprocess steps in teneto.derive
    """

    # Create report directory
    if not os.path.exists(sdir):
        os.makedirs(sdir)

    # Add a slash to file directory if not included to avoid DirNameFleName
    # instead of DirName/FileName being creaated
    if sdir[-1] != '/':
        sdir += '/'

    report_html = '<html><body>'

    if 'method' in report.keys():

        report_html += "<h1>Method: " + report['method'] + "</h1><p>"

        for i in report[report['method']]:

            if i == 'taper_window':

                fig, ax = plt.subplots(1)
                ax.plot(report[report['method']]['taper_window'],
                        report[report['method']]['taper'])
                ax.set_xlabel('Window (time). 0 in middle of window.')
                ax.set_title(
                    'Taper from ' + report[report['method']]['distribution'] + ' distribution (PDF).')
                fig.savefig(sdir + 'taper.png')

                report_html += "<img src='./taper.png' width=500>" + "<p>"

            else:

                report_html += "- <b>" + i + "</b>: " + \
                    str(report[report['method']][i]) + "<br>"

    if 'postprocess' in report.keys():

        report_html += "<p><h2>Postprocessing:</h2><p>"

        report_html += "<b>Pipeline: </b>"

        for i in report['postprocess']:

            report_html += " " + i + ","

        for i in report['postprocess']:

            report_html += "<p><h3>" + i + "</h3><p>"

            for j in report[i]:

                if j == 'lambda':

                    report_html += "- <b>" + j + "</b>: " + "<br>"

                    lambda_val = np.array(report['boxcox']['lambda'])
                    fig, ax = plt.subplots(1)
                    ax.hist(lambda_val[:, -1])
                    ax.set_xlabel('lambda')
                    ax.set_ylabel('frequency')
                    ax.set_title('Histogram of lambda parameter')
                    fig.savefig(sdir + 'boxcox_lambda.png')

                    report_html += "<img src='./boxcox_lambda.png' width=500>" + "<p>"
                    report_html += "Data located in " + sdir + "boxcox_lambda.csv <p>"

                    np.savetxt(sdir + "boxcox_lambda.csv", lambda_val, delimiter=",")

                else:

                    report_html += "- <b>" + j + "</b>: " + \
                        str(report[i][j]) + "<br>"

    report_html += '</body></html>'

    with open(sdir + report_name, 'w') as file:
        file.write(report_html)
    file.close()



import numpy as np
import teneto.networkmeasures
import teneto.utils
import copy

def add_node(netObj,n=1,nLab=''):
    """
    Adds a node to the end of network.
        :n: How many nodes to add. Default 1. Can be any integer.
        :nLab: list of labels for added nodes. len(nLab)==n
    """
    #If input is string make to list. Repeat list if n>1 and nLab is string.
    if nLab != '' and isinstance(nLab,str):
        nLab = [nLab] * n

    #Check that number of nLabs is compatible with number of new nodes. Add new nLabs
    if nLab!='' and n!=len(nLab):
        raise ValueError('Specified node labels need to each number of new nodes')
    elif netObj.contact['nLabs']!='' and nLab == '':
        netObj.contact['nLabs'] = netObj.contact['nLabs'] + ['UnnamedNode'] * n
    elif nLab != '':
        netObj.contact['nLabs'] = netObj.contact['nLabs'] + nLab

    #Add new
    netObj.contact['netshape']=tuple([netObj.contact['netshape'][0]+n, netObj.contact['netshape'][1]+n, netObj.contact['netshape'][2]])
    return netObj


def add_time(netObj,t=1):
    """
    Adds a time-point to the end of network.
        :t: How many time-points to add. Default 1. Can be any integer.
    """
    netObj.contact['netshape']=tuple([netObj.contact['netshape'][0], netObj.contact['netshape'][1], netObj.contact['netshape'][2]+t])
    return netObj

def add_edge(netObj,e,dimRefNode='index'):
    """
    Adds a an edge.

    :e: nx3 (bu,bd) or nx4 (wu,wd)array. Where n is the number of edges to be added
        :for binary: columns are [node,node,time]
        :for wegithed: columns are [node,node,time,value]

    :dimRefNode: reference node 'index' or 'nLabs'
    """
    e=np.array(e)
    if netObj.contact['nettype'][0]=='b' and e.shape[-1]!=3:
        raise ValueError('For binary networks, e must have len of 3 (node,node,time)')
    elif netObj.contact['nettype'][0]=='w' and e.shape[-1]!=4:
        raise ValueError('For weighted networks, e must have len of 4 (node,node,time,value)')
    for edge in e:
        if edge[0] >= netObj.contact['netshape'][0] or edge[1] >= netObj.contact['netshape'][1]:
            raise ValueError('Nodes must be defined before nodes passes through edge')
        if edge[2] >= netObj.contact['netshape'][2]:
            raise ValueError('Time index must be defind before edge defind')
        if np.sum(np.sum(netObj.contact['contacts']-edge==0,axis=1)==3)>0:
            raise ValueError('Edge already exists')
        netObj.contact['contacts']=np.vstack([netObj.contact['contacts'],edge[0:3]])
        if netObj.contact['nettype'][0]=='w':
            netObj.contact['values']=np.vstack([netObj.contact['values'],edge[3]])
    return netObj

def add_graphlet(netObj,G=[]):
    """
    Adds a graphlet.

    :G: 2 or 3 dimensional graphlet.

    """
    if G==[]:
        netobj = add_time(netObj)
    elif netObj.contact['netshape'][0]!=G.shape[0] or netObj.contact['netshape'][1]!=G.shape[1]:
        raise ValueError('Graphlets added must be of same size')
    else:
        C = teneto.utils.graphlet2contact(G)
        netObj = add_contact(netObj,C)
    return netObj

def add_contact(netObj,C,time='append'):
    """
    Adds a graphlet.

    :C: contact of the same size.
    :time: 'append' means all contacts are added to the end. 'asis', contacts are added as they are.

    """
    if netObj.contact['netshape'][0]!=C['netshape'][0]:
        raise ValueError('Contact shapes added must be of same size')
    if time == 'append':
        C = copy.deepcopy(C)
        C['contacts'][:,-1] += netObj.contact['netshape'][-1]
        netObj.contact['netshape']=(netObj.contact['netshape'][0],netObj.contact['netshape'][1],netObj.contact['netshape'][-1]+C['netshape'][-1])
        netObj.contact['contacts']=np.array(np.vstack([netObj.contact['contacts'],C['contacts']]))

    return netObj


class NewTemporalNetwork:

    def __init__(self):

        self.contact = []


    def generate_rand_binomial(self,size,p,nettype='bu',initialize='zero',netinfo=None):
        if netinfo != None:
            netinfo = dict(netinfo)
        self.contact= teneto.generatenetwork.rand_binomial(size,p,'contact',nettype,initialize,netinfo)


    def get_shape(self,):
        print(self.contact['netshape'])

    def get_nettype(self,):
        print(self.contact['nettype'])

    def get_graphlet_representation(self,):
        return teneto.utils.contact2graphlet(self.contact)

    def from_graphlet_representation(self,G):
        self.contact = teneto.utils.graphlet2contact(G)

    def CreateHyperGraph(self):
        return teneto.HyperGraph(self)

    def add_node(self,n=1,nLab=''):
        self = add_node(self,n,nLab)

    def add_time(self,t=1):
        self = add_time(self,t)

    def add_edge(self,e,dimRefNode='index'):
        self = add_edge(self,e)

    def add_graphlet(self,G=[]):
        self = add_graphlet(self,G)

    def add_contact(self,C,time='append'):
        self = add_contact(self,C,time=time)

# class HyperGraph:
#
#     def __init__(self):
#
#         self.contact = []
#
#     def get_shape(self,):
#         print(self.contact['netshape'])
#
#     def get_nettype(self,):
#         print(self.contact['nettype'])
#
#     def get_graphlet_representation(self,):
#         return teneto.utils.contact2graphlet(self.contact)
#
#     def from_graphlet_representation(self,G):
#         self.contact = teneto.utils.graphlet2contact(G)
#
#     def add_node(self,n=1,nLab=''):
#         self = add_node(self,n,nLab)
#
#     def add_edge(self,e,dimRefNode='index'):
#         self = add_edge(self,e)



"""
Network mesures: temporal degree centrality
"""

import numpy as np
import teneto.utils as utils


def temporal_degree_centrality(net, axis=0, calc='avg', subnet=None, decay=None):
    """

    temporal degree of network. Sum of all connections each node has through time.

    **PARAMETERS**

    :net: temporal network input (graphlet or contact).

        :nettype: 'bu', 'bd', 'wu', 'wd'

    :axis: Dimension that is returned 0 or 1 (default 0).
        Note, only relevant for directed networks.
        i.e. if 0, node i has Aijt summed over j and t.
        and if 1, node j has Aijt summed over i and t.

    :calc: 'avg' (returns temporal degree centrality (a 1xnode vector))
     or 'time' (returns a node x time matrix).
     i.e. 'time' returns static degree centrality per time-point.

    :subnet: None (default) or Nx1 vector of subnetwork assignment.
    This returns a "centrality" per subnetwork instead of per node.

    :decay: if calc = 'time', then decay is possible where the centrality of
    the previous time point is carried over to the next time point but decays
    at a value of $e^decay$ such that $D_d(t+1) = e^{-decay}D_d(t) + D(t+1)$. If
    decay is 0 then the final D will equal D when calc='avg', if decay = inf
    then this will equal calc='time'.

    **OUTPUT**

    :D: temporal degree centrality (nodal measure)

        :format: 1d numpy array (or 2d if calc = 'time')

    **SEE ALSO**

    - *temporal_closeness_centrality*

    """

    # Get input in right format
    net, netinfo = utils.process_input(net, ['C', 'G', 'TO'])

    # sum sum net
    if calc == 'time' and not subnet:
        tdeg = np.squeeze(np.sum(net, axis=axis))
    elif calc != 'time' and not subnet:
        tdeg = np.sum(
            np.sum(net, axis=2), axis=axis)
    elif calc == 'time' and subnet:
        unique_subnet = np.unique(subnet)
        tdeg_subnet = [np.sum(np.sum(net[subnet == s1, :, :][:, subnet == s2, :], axis=1), axis=0)
                       for s1 in unique_subnet for s2 in unique_subnet]


        tdeg = np.array(tdeg_subnet)
        tdeg = np.reshape(tdeg, [len(np.unique(subnet)), len(
            np.unique(subnet)), net.shape[-1]])
        # Divide diagonal by 2 if undirected to correct for edges being present twice
        if netinfo['nettype'][1] == 'u':
            for s in range(tdeg.shape[0]):
                tdeg[s,s,:] = tdeg[s,s,:]/2
    else:
        raise ValueError("invalid calc argument")
    if decay and calc=='time':
        #Reshape so that time is first dimensions
        tdeg = tdeg.transpose(np.hstack([len(tdeg.shape)-1,np.arange(len(tdeg.shape)-1)]))
        for n in range(1,tdeg.shape[-1]):
            tdeg[n] = np.exp(-decay)*tdeg[n-1] + tdeg[n]
        tdeg = tdeg.transpose(np.hstack([np.arange(1,len(tdeg.shape)),0]))
    elif decay:
        print('WARNING: decay cannot be applied unless calc=time, ignoring decay')

    return tdeg



"""
General setup for module
"""

from setuptools import setup, find_packages

setup(name='teneto',
      version='0.3.0',
      python_requires='>3.5',
      install_requires=[  
			'nilearn>=0.4.0',
			'statsmodels>=0.8.0',
			'pybids>=0.4.2',
      'louvain>=0.6.1',
      'python-igraph>=0.6.1'],
      description='Temporal network tools',
      packages=find_packages(),
      author='William Hedley Thompson',
      author_email='hedley@startmail.com',
      url='https://www.github.com/wiheto/teneto',
      download_url='https://github.com/wiheto/teneto/archive/0.2.tar.gz',
      package_data={'':['./teneto/data']},
      include_package_data = True)



"""
networkmeasures.temporal_closeness_centrality
"""

import numpy as np
from teneto.networkmeasures.shortest_temporal_path import shortest_temporal_path


def temporal_closeness_centrality(data):
    '''
    returns temporal closeness centrality per node.
    As temporalPaths only works with binary undirected edges at the moment,
     this is required for temporal closeness centrality.

    **PARAMETERS**

    :data: This is either:

        :netIn: temporal network input (graphlet or contact).

            :nettype: 'bu'

        :paths: Dictionary of paths (output of shortest_temporal_path).


    **OUTPUT**

    :C: temporal closness centrality (nodal measure)

        :format: 1d numpy array

    **See Also**

    - *temporalPaths*
    - *temporalDegree*

    **History**

    Modified - Dec 2016, WHT (documentation, cleanup)
    Created - Nov 2016, WHT

    '''

    pathdata = 0  # are shortest paths calculated
    if isinstance(data, dict):
        # This could be done better
        if [k for k in list(data.keys()) if k == 'paths'] == ['paths']:
            pathdata = 1
    # if shortest paths are not calculated, calculate them
    if pathdata == 0:
        data = shortest_temporal_path(data)

    closeness = np.nansum(
        1 / np.nanmean(data['paths'], axis=2), axis=1) / (data['paths'].shape[1] - 1)

    return closeness




# #!/usr/bin/env python3
#
"""Teneto is a module with tools for analyzing network patterns in time."""
#
__author__ = "William Hedley Thompson (wiheto)"
__version__ = "0.3.0"
#
import teneto.utils as utils
import teneto.networkmeasures as networkmeasures
import teneto.plot as plot
import teneto.generatenetwork as generatenetwork
import teneto.derive as derive
import teneto.misc as misc
import teneto.trajectory as trajectory
import teneto.communitydetection as communitydetection
from teneto.classes import TenetoBIDS
#del misc.teneto
#del communitydetection.static.modularity_based_clustering

#del measures



# Main function to draw a slice_graph
import matplotlib.pyplot as plt
import numpy as np
from teneto.utils import *


def slice_plot(netIn, ax, nLabs='', tLabs='', timeunit='', linestyle='k-', cmap=None, nodesize=100):
    '''

    Fuction draws "slice graph" and exports axis handles


    **PARAMETERS**

    :netIn: temporal network input (graphlet or contact)
    :ax: matplotlib figure handles.
    :nLabs: nodes labels. List of strings.
    :tLabs: labels of dimension Graph is expressed across. List of strings.
    :timeunit: unit time axis is in.
    :linestyle: line style of Bezier curves.
    :nodesize: size of nodes


    **OUTPUT**

    :ax: axis handle of slice graph


    **SEE ALSO**

    - *circle_plot*
    - *graphlet_stack_plot*


    **HISTORY**

    :modified: Dec 2016, WHT (documentation, improvments)
    :created: Sept 2016, WHT

    '''
    # Get input type (C or G)
    inputType = checkInput(netIn)
    nettype = 'xx'
    # Convert C representation to G

    if inputType == 'G':
        cfg = {}
        netIn = graphlet2contact(netIn)
        inputType = 'C'
    edgeList = [tuple(np.array(e[0:2]) + e[2] * netIn['netshape'][0])
                for e in netIn['contacts']]

    if nLabs != '' and len(nLabs) == netIn['netshape'][0]:
        pass
    elif nLabs != '' and len(nLabs) != netIn['netshape'][0]:
        raise ValueError('specified node label length does not match netshape')
    elif nLabs == '' and netIn['nLabs'] == '':
        nLabs = np.arange(1, netIn['netshape'][0] + 1)
    else:
        nLabs = netIn['nLabs']

    if tLabs != '' and len(tLabs) == netIn['netshape'][-1]:
        pass
    elif tLabs != '' and len(tLabs) != netIn['netshape'][-1]:
        raise ValueError('specified time label length does not match netshape')
    elif tLabs == '' and str(netIn['t0']) == '':
        tLabs = np.arange(1, netIn['netshape'][-1] + 1)
    else:
        tLabs = np.arange(netIn['t0'], netIn['Fs'] *
                          netIn['netshape'][-1] + netIn['t0'], netIn['Fs'])

    if timeunit == '':
        timeunit = netIn['timeunit']

    timeNum = len(tLabs)
    nodeNum = len(nLabs)
    pos = []
    posy = np.tile(list(range(0, nodeNum)), timeNum)
    posx = np.repeat(list(range(0, timeNum)), nodeNum)

    node_plot_attr = {}
    if cmap:
        node_plot_attr['cmap'] = cmap


    # plt.plot(points)
    # Draw Bezier vectors around egde positions
    for edge in edgeList:
        bvx, bvy = bezier_points(
            (posx[edge[0]], posy[edge[0]]), (posx[edge[1]], posy[edge[1]]), nodeNum, 20)
        ax.plot(bvx, bvy, linestyle)
    ax.set_yticks(range(0, len(nLabs)))
    ax.set_xticks(range(0, len(tLabs)))
    ax.set_yticklabels(nLabs)
    ax.set_xticklabels(tLabs)
    ax.grid()
    ax.set_frame_on(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()
    ax.set_xlim([min(posx) - 1, max(posx) + 1])
    ax.set_ylim([min(posy) - 1, max(posy) + 1])
    ax.scatter(posx, posy, s=nodesize, c=posy, zorder=10, **node_plot_attr)
    if timeunit != '':
        timeunit = ' (' + timeunit + ')'
    ax.set_xlabel('Time' + timeunit)

    return ax


# Following 3 Function that draw vertical curved lines from around points.
# p1 nad p2 are start and end trupes (x,y coords) and pointN is the resolution of the points
# negxLim tries to restrain how far back along the x axis the bend can go.
def bezier_points(p1, p2, negxLim, pointN):
    ts = [t / pointN for t in range(pointN + 1)]
    d = p1[0] - (max(p1[1], p2[1]) - min(p1[1], p2[1])) / negxLim
    bezier = make_bezier([p1, (d, p1[1]), (d, p2[1]), p2])
    points = bezier(ts)
    bvx = [i[0] for i in points]
    bvy = [i[1] for i in points]
    return bvx, bvy


# These two functions originated from the plot.ly's documentation for python API.
# They create points along a curve.
def make_bezier(xys):
    # xys should be a sequence of 2-tuples (Bezier control points)
    n = len(xys)
    combinations = pascal_row(n - 1)

    def bezier(ts):
        # This uses the generalized formula for bezier curves
        # http://en.wikipedia.org/wiki/B%C3%A9zier_curve#Generalization
        result = []
        for t in ts:
            tpowers = (t**i for i in range(n))
            upowers = reversed([(1 - t)**i for i in range(n)])
            coefs = [c * a * b for c, a,
                     b in zip(combinations, tpowers, upowers)]
            result.append(
                tuple(sum([coef * p for coef, p in zip(coefs, ps)]) for ps in zip(*xys)))
        return result
    return bezier


def pascal_row(n):
    # This returns the nth row of Pascal's Triangle
    result = [1]
    x, numerator = 1, n
    for denominator in range(1, n // 2 + 1):
        # print(numerator,denominator,x)
        x *= numerator
        x /= denominator
        result.append(x)
        numerator -= 1
    if n & 1 == 0:
        # n is even
        result.extend(reversed(result[:-1]))
    else:
        result.extend(reversed(result))
    return result



import teneto
import numpy as np


def rdp(datin,delta=1,report=10,quiet=True):


    # This needs to be added to utils for trajectory detection
    # T will be for trajectory or timeseries data, (roi x time).
    datin, datinfo = teneto.utils.process_input(datin, ['C', 'G', 'TO', 'T'])
    # if network, then make to roi,time shape. If T, nothing needs to be done.
    if len(datin.shape) == 3:
        ind = np.triu_indices(datin.shape[0],k=1)
        datin = datin[ind[0],ind[1],:]
        index_out = np.array([ind[0],ind[1]]).transpose()
    else:
        index_out = np.arange(0,datin.shape[0])

    #Interprets data to be rois x time.
    datin = np.array(datin,ndmin=2)
    s = 0
    e = datin.shape[-1]

    # Create straight line between start and end point
    trajectory = teneto.utils.create_traj_ranges(datin[:,s],datin[:,e-1],e-s)

    # Create lists of trajectories
    trajectory_points = [np.array([s,e-1]) for n in np.arange(datin.shape[0])]
    # Preset some outputs
    reduction = np.zeros(len(trajectory_points))
    error = np.zeros(len(trajectory_points))
    amp_weighted_error = np.zeros(len(trajectory_points))
    # Stopping condition and round preset
    stop_cond = 0
    round_count = 0
    while stop_cond < datin.shape[0]:
        traj_data_diff = np.abs(datin-trajectory)
        ind = np.argmax(traj_data_diff,axis=1)

        for i,ix in enumerate(ind):
            ind_bool = traj_data_diff[i,ix] > delta
            if ind_bool:
                # Get trajectory breaking point indicies
                traj_ind = np.searchsorted(trajectory_points[i],ix,side='left')
                traj_start = trajectory_points[i][traj_ind-1]
                traj_end = trajectory_points[i][traj_ind]
                # Make new trajectories
                r1 = np.arange(traj_start,ix+1)
                trajectory[i,r1] = teneto.utils.create_traj_ranges(trajectory[i,traj_start],datin[i,ix],len(r1))
                r2 = np.arange(ix,traj_end+1)
                trajectory[i,r2] = teneto.utils.create_traj_ranges(datin[i,ix],trajectory[i,traj_end],len(r2))
                # Add new point
                trajectory_points[i] = np.insert(trajectory_points[i],traj_ind,ix)
            else:
                stop_cond += 1

        round_count += 1
        if np.remainder(round_count,report) == 0 and not quiet:
            print('After round: ' + str(round_count) + ', ' + str(np.round((stop_cond/datin.shape[0])*100,2)) + '% completed data reduction')

    for i in np.arange(datin.shape[0]):
        reduction[i] = float(1-len(trajectory_points[i])/datin.shape[1])
        error[i] = float(np.mean(traj_data_diff[i]))
        amp_weighted_error[i] = float(np.mean(np.abs(datin[i,:])*traj_data_diff[i]))

    # Construct ouput dictionary
    traj = datinfo
    traj['trajectory'] = trajectory
    traj['trajectory_points'] = trajectory_points
    traj['reduction'] = reduction
    traj['error'] = error
    traj['weighted_error'] = amp_weighted_error
    traj['nettype'] = 't'
    traj['index'] = index_out
    traj.pop('inputtype')

    return traj
